/* NAZA GPS ENCODER V0.2 alpha
// Code by Luke Tanti 2015
// Thansk Pawelsky for the Decoder!!
// Thansk Ciskje and Andreaontheweb for help!!
*/

#include <Wire.h>
#include <AltSoftSerial.h>
#include <I2C.h>

#define HMC5883L  0x1E 

AltSoftSerial altSerial;

PROGMEM  prog_uchar confUBX0[] = {0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0x01, 0x02, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x17, 0xCD};//POSLLH
PROGMEM  prog_uchar confUBX1[] = {0xB5, 0x62, 0x06, 0x01, 0x02, 0x00, 0x01, 0x02, 0x0C, 0x35};//SAVE POSLLH
PROGMEM  prog_uchar confUBX2[] = {0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0x01, 0x04, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x19, 0xDB};//DOP
PROGMEM  prog_uchar confUBX3[] = {0xB5, 0x62, 0x06, 0x01, 0x02, 0x00, 0x01, 0x04, 0x0E, 0x37};//SAVE DOP
PROGMEM  prog_uchar confUBX4[] = {0xB5, 0x62, 0x06, 0x01, 0x08, 0x00, 0x01, 0x21, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x36, 0xA6};//TIMEUTC
PROGMEM  prog_uchar confUBX5[] = {0xB5, 0x62, 0x06, 0x01, 0x02, 0x00, 0x01, 0x21, 0x2B, 0x54};//SAVE TIMEUTC
PROGMEM  prog_uchar confUBX6[] = {0xB5, 0x62, 0x06, 0x08, 0x06, 0x00, 0xC8, 0x00, 0x01, 0x00, 0x01, 0x00, 0xDE, 0x6A};//Change message frequency to 5Hz
PROGMEM  prog_uchar confUBX7[] = {0xB5, 0x62, 0x06, 0x08, 0x00, 0x00, 0x0E, 0x30};//Save Rate

uint8_t moduleVer[18] = {0x55, 0xAA, 0x30, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x02, 0x01, 0x00, 0x06, 0x01, 0x01, 0x01, 0x06, 0x4E, 0x80};
char GPSdata[60];

unsigned long time0x30 = 0;

boolean gotUBX = false;

char compassValues[8];
uint8_t noMask = 0x00;
uint8_t CK_A = 0, CK_B = 0;
uint8_t mask = 0x00;

uint16_t sequenceNumber = 0;
uint16_t datetimeL = 0;
uint16_t datetimeH = 0;
uint16_t year = 0;
boolean TIMEUTC = 0;
boolean POSLLH = 0;
boolean VELNED = 0;
boolean DOP = 0;
boolean SOL = 0;
boolean swap = 0;

//Get UBX Message
int gpsStep = 0;
uint8_t UBX_id = 0;
uint8_t UBX_class = 0;
uint8_t meslenL = 0;
uint8_t meslenH = 0;
int16_t meslen = 0;
int count = 0;
uint8_t CK_AU = 0, CK_BU = 0;
uint8_t message[60];
boolean temp = 0;

void sendUBX(prog_uchar *array, uint8_t length)
{
  for (uint8_t k=0;k<length;k++)
  {
    altSerial.write(pgm_read_word_near(array + k));
    delay(2);
  }
  altSerial.println();
  delay(2);
}

void initUBX(void)
{
  sendUBX(&confUBX0[0], sizeof(confUBX0));
  sendUBX(&confUBX1[0], sizeof(confUBX1));
  sendUBX(&confUBX2[0], sizeof(confUBX2));
  sendUBX(&confUBX3[0], sizeof(confUBX3));
  sendUBX(&confUBX4[0], sizeof(confUBX4));
  sendUBX(&confUBX5[0], sizeof(confUBX5));
  sendUBX(&confUBX6[0], sizeof(confUBX6));
  sendUBX(&confUBX7[0], sizeof(confUBX7));
}

void sendVersion()
{
  for (int i=0;i<18;i++)
  {
    Serial.write(moduleVer[i]);
  }
}

void sendHMC5883L()
{
  boolean temp;
  I2c.read(HMC5883L,0x03,6); //read 6 bytes (x,y,z) from the device
  int x = 0;
  int y = 0;
  x = I2c.receive() << 8;
  x |= I2c.receive(); 
  compassValues[5] = x >> 8;
  compassValues[4] = x;
  compassValues[7] = I2c.receive();  // Z
  compassValues[6] = I2c.receive();  // Z
  y = I2c.receive() << 8;
  y |= I2c.receive(); 
  y = -y;
  compassValues[3] = y >> 8;
  compassValues[2] = y;
  
  //Generate Mask using XOR
  noMask = compassValues[6];
  mask = (bitRead(compassValues[6], 0)) ^ (bitRead(compassValues[6], 4));
  temp = (bitRead(compassValues[6], 1)) ^ (bitRead(compassValues[6], 5)); 
  mask |= (temp << 1);
  temp = (bitRead(compassValues[6], 2)) ^ (bitRead(compassValues[6], 6)); 
  mask |= (temp << 2);
  temp = (bitRead(compassValues[6], 3)) ^ (bitRead(compassValues[6], 7) ^ (bitRead(compassValues[6], 0))); 
  mask |= (temp << 3);
  mask |= bitRead(compassValues[6], 1) << 4;
  mask |= bitRead(compassValues[6], 2) << 5;
  mask |= bitRead(compassValues[6], 3) << 6;
  temp = (bitRead(compassValues[6], 4)) ^ (bitRead(compassValues[6], 0)); 
  mask |= (temp << 7);
  //
  compassValues[0] = 0x20 ^ mask;
  compassValues[1] = 0x06 ^ mask;
  //
  Serial.write(0x55);
  //Serial.flush();
  Serial.write(0xAA);
  //Serial.flush();
  for(int i=0;i<8;i++)
  {
    if (i == 6)
    {
      Serial.write(noMask);
      //Serial.flush();
      CK_A = CK_A + noMask;
      CK_B = CK_B + CK_A;
    }
    else
    {
      Serial.write(compassValues[i] ^ mask);
      //Serial.flush();
      CK_A = CK_A + (compassValues[i] ^ mask);
      CK_B = CK_B + CK_A;
    }
  }
  Serial.write(CK_A);
  //Serial.flush();
  Serial.write(CK_B);
  //Serial.flush();
  CK_A = 0;
  CK_B = 0;
  mask = 0x00;
}

// uBlox checksum algorithm
void ubx_checksum(byte ubx_data)
{
  CK_A+=ubx_data;
  CK_B+=CK_A;
}

void clearVars(void)
{
  gpsStep = 0;
  UBX_id = 0;
  UBX_class = 0;
  meslen = 0;
  count = 0;
  CK_AU = 0;
  CK_BU = 0;
  CK_A = 0;
  CK_B = 0;
  memset(message, 0, 60);
}

void sendGPS(void)
{
  gotUBX = true;
  CK_A = 0;
  CK_B = 0;
  GPSdata[0] = 0x10;
  GPSdata[1] = 0x3A;
  if (TIMEUTC == 1 && POSLLH == 1 && VELNED == 1 && DOP == 1 && SOL == 1)
  {
    if (GPSdata[52] == 0x02 || GPSdata[52] == 0x03)  //Check for Fix
    {
      sequenceNumber++;
      GPSdata[58] = sequenceNumber & 0xFF;
      GPSdata[59] = (sequenceNumber & 0xFF00) >> 8;
    }
    else
    {
      if (swap == true)
      {
        GPSdata[58] = sequenceNumber & 0xFF;
        GPSdata[59] = (sequenceNumber & 0xFF00) >> 8;
        swap = false;
      }
      if (swap == false)
      {
        GPSdata[59] = sequenceNumber & 0xFF;
        GPSdata[58] = (sequenceNumber & 0xFF00) >> 8;
        swap = true;
      }
    }
    //Generate Mask
    mask = (bitRead(GPSdata[50], 0)) ^ (bitRead(GPSdata[58], 4));
    temp = (bitRead(GPSdata[50], 1)) ^ (bitRead(GPSdata[58], 5)); 
    mask |= (temp << 1);
    temp = (bitRead(GPSdata[50], 2)) ^ (bitRead(GPSdata[58], 6)); 
    mask |= (temp << 2);
    temp = (bitRead(GPSdata[50], 3)) ^ (bitRead(GPSdata[58], 7) ^ (bitRead(GPSdata[50], 0))); 
    mask |= (temp << 3);
    mask |= bitRead(GPSdata[50], 1) << 4;
    mask |= bitRead(GPSdata[50], 2) << 5;
    mask |= bitRead(GPSdata[50], 3) << 6;
    temp = (bitRead(GPSdata[50], 0)) ^ (bitRead(GPSdata[58], 4)); 
    mask |= (temp << 7);
    //Mask unmaskable variables so that the mask cancels out
    GPSdata[0]^= mask;
    GPSdata[1]^= mask;
    GPSdata[50]^= mask;
    GPSdata[51]^= mask;
    GPSdata[58]^= mask;
    GPSdata[59]^= mask;
    //
    Serial.write(0x55);
    //Serial.flush();
    Serial.write(0xAA);
    //Serial.flush();
    for(int i=0;i<60;i++)
    {
        Serial.write(GPSdata[i] ^ mask);
        //Serial.flush();
        CK_A = CK_A + (GPSdata[i] ^ mask);
        CK_B = CK_B + CK_A;
    }
    Serial.write(CK_A);
    //Serial.flush();
    Serial.write(CK_B);
    //Serial.flush();
    //
    TIMEUTC = 0;
    POSLLH = 0;
    VELNED = 0;
    DOP = 0;
    SOL = 0;
    CK_A = 0;
    CK_B = 0;
    mask = 0;
  }
}

void decodeUBX(void)
{
  if (UBX_class == 0x01)
  {
    switch(UBX_id)
    {
      case 0x21:  //NAV-TIMEUTC
      {        
        datetimeL = (message[18] & 0x3F);             // Second
        datetimeL |= (message[17] & 0x3F) << 6;       // Minute
        datetimeL |= (message[16] & 0x0F) << 12;      // Hour
        if ((message[16] & 0x0F) > 7)
          datetimeH |= ((message[15]-1) & 0x1F);      // Day
        else
          datetimeH |= (message[15] & 0x1F);          // Day     
        datetimeH |= (message[14] & 0x0F) << 5;       // Month
        year = message[12];
        year |= message[13] << 8;
        year-= 2000;
        datetimeH |= (year & 0x7F)  << 9;             // Year
        GPSdata[2] = datetimeL;
        GPSdata[3] = datetimeL >> 8;
        GPSdata[4] = datetimeH;
        GPSdata[5] = datetimeH >> 8;       
        TIMEUTC = true;
        break;
      }
      case 0x02:  //ID NAV-POSLLH
      {
        GPSdata[6] = message[4];    //LON-0
        GPSdata[7] = message[5];    //LON-1
        GPSdata[8] = message[6];    //LON-2
        GPSdata[9] = message[7];    //LON-3
        GPSdata[10] = message[8];   //LAT-0
        GPSdata[11] = message[9];   //LAT-1
        GPSdata[12] = message[10];  //LAT-2
        GPSdata[13] = message[11];  //LAT-3
        GPSdata[14] = message[16];  //ALT-0
        GPSdata[15] = message[17];  //ALT-1
        GPSdata[16] = message[18];  //ALT-2
        GPSdata[17] = message[19];  //ALT-3  
        GPSdata[18] = message[20];  //HAC-0 
        GPSdata[19] = message[21];  //HAC-1
        GPSdata[20] = message[22];  //HAC-2
        GPSdata[21] = message[23];  //HAC-3
        GPSdata[22] = message[24];  //VAC-0
        GPSdata[23] = message[25];  //VAC-1
        GPSdata[24] = message[26];  //VAC-2
        GPSdata[25] = message[27];  //VAC-3
        GPSdata[26] = 0x00;         //XX-0
        GPSdata[27] = 0x00;         //XX-1
        GPSdata[28] = 0x00;         //XX-2
        GPSdata[29] = 0x00;         //XX-3
        POSLLH = true;
        break;
      }
      case 0x12:  //NAV-VELNED
      {
        GPSdata[30] = message[4];   //NVEL-0
        GPSdata[31] = message[5];   //NVEL-1
        GPSdata[32] = message[6];   //NVEL-2
        GPSdata[33] = message[7];   //NVEL-3
        GPSdata[34] = message[8];   //EVEL-0
        GPSdata[35] = message[9];   //EVEL-1
        GPSdata[36] = message[10];  //EVEL-2
        GPSdata[37] = message[11];  //EVEL-3
        GPSdata[38] = message[12];  //DVEL-0
        GPSdata[39] = message[13];  //DVEL-1
        GPSdata[40] = message[14];  //DVEL-2
        GPSdata[41] = message[15];  //DVEL-3
        VELNED = true;
        break;
      }
      case 0x04:  //NAV-DOP
      {
        GPSdata[42] = message[6];   //PDOP-0
        GPSdata[43] = message[7];   //PDOP-1
        GPSdata[44] = message[10];  //VDOP-0
        GPSdata[45] = message[11];  //VDOP-1
        GPSdata[46] = message[14];  //NDOP-0
        GPSdata[47] = message[15];  //NDOP-1
        GPSdata[48] = message[16];  //EDOP-0
        GPSdata[49] = message[17];  //EDOP-1     
        DOP = true;  
        break;
      }
      case 0x06:  // NAV-SOL
      {
        GPSdata[50] = message[47];  //Number of SVs used in Nav Solution
        GPSdata[51] = 0x00;         //not XORed      <--------------------------------------
        GPSdata[52] = message[10];  //Fix type
        GPSdata[53] = 0x00;
        GPSdata[54] = message[11];  //Fix Status Flags
        GPSdata[55] = 0x00;
        GPSdata[56] = 0x00;
        GPSdata[57] = 0x00;
        SOL = true;
        break;
      }
      case 0x30:
      {
        break;
      }
      case 0x03:
      {
        break;
      }
      case 0x22:
      {
        break;
      }
    }
  }
  sendGPS();
}

void getUBX(void)
{  
  if (altSerial.available())
  {
    uint8_t c = altSerial.read();
    switch (gpsStep)
    {
      case 0:
      {
        if (c == 0xB5)
          gpsStep++;
        break;
      }
      case 1:
      {
        if (c == 0x62)
          gpsStep++;
        else
          gpsStep = 0;
        break;
      }
      case 2:
      {
        UBX_class = c;
        ubx_checksum(c);
        gpsStep++;
        break;
      }
      case 3:
      {
        UBX_id = c;
        ubx_checksum(c);
        gpsStep++;
        break;
      }
      case 4:
      {
        meslenL = c;
        ubx_checksum(c);
        gpsStep++;
        break;
      }
      case 5:
      {
        meslenH = c;
        ubx_checksum(c);
        gpsStep++;
        meslen = 0xFF & meslenL;
        meslen |= meslenH << 8;
        count = 0;
        break;
      }
      case 6:
      {
        message[count] = c;
        ubx_checksum(c);
        count++;
        if (count == meslen)
        {
          gpsStep++;
          count = 0;
        }
        break;        
      }
      case 7:
      {
        CK_AU = c;
        gpsStep++;
        break;
      }
      case 8:
      {
        CK_BU = c;
        if (CK_A == CK_AU && CK_B == CK_BU)
        {    
          decodeUBX();
        }
        clearVars();
        break;
      }
    }
  }  
}

void setup()
{
  Serial.begin(115200);
  altSerial.begin(57600);
  I2c.begin();
  I2c.write(HMC5883L,0x00,0x18); //configure device speed to 75HZ
  delay(1);
  I2c.write(HMC5883L,0x02,0x00); //configure device for continuous mode
  delay(1);
  initUBX();
  pinMode(13,OUTPUT);
}

void loop()
{
  getUBX();
  if (gotUBX == true)
  {
    sendHMC5883L();
    digitalWrite(13, digitalRead(13) ^ 1);
    if ((millis() - time0x30) >= 2000)
    {
      sendVersion();
      time0x30 = millis();
    }
    gotUBX = false;
  }
}

